/* (C) 2018 King Kevin <kingkevin@cuvoodoo.info>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
/* Library to decode IR NEC transmissions.
 * More information about the protocol at http://www.sbprojects.com/knowledge/ir/nec.php
 */
#include <stdint.h> // Standard Integer Types
#include <stdlib.h> // Standard Library
#include <stdbool.h> // Standard Boolean Type

#include "main.h" // global definitions
#include "ir_nec.h" // own definitions

// mark and space durations to encode NEC bursts (in increase order)
const static uint16_t MARKS[2]={560, 9000}; // possible mark duration in µs { bit, start }
const static uint16_t SPACES[4]={1120-560, 2250-560, 2250, 4500}; // possible space duration in µs  { 0, 1 , repeat, start}

/* convert mark/space pulse times to closest mark/space NEC type
 * @param[in] burst the mark/space pulses (mark on even index, space on odd index)
 * @param[in] pulses numbers of pulses
 * @param[out] nec the NEC mark and space types (index of MARKS or SPACE)
 */
static void burst2nec(const uint16_t* burst, uint8_t pulses, uint8_t* nec)
{
	for (uint8_t i = 0; i < pulses; i++) { // go through the pulses in the burst
		if (0 == burst[i]) { // end of burst
			return;
		}
		nec[i] = 0; // default mark/space index (to be updated)
		if (0 == i % 2) { // even index is for mark bursts
			// find the nearest NEC mark time
			for (uint8_t mark = 1; mark < COUNT_OF(MARKS); mark++) {
				if (burst[i] > ((MARKS[mark-1] + MARKS[mark]) / 2)) {
					nec[i] = mark;
				} else {
					break;
				}
			}
		} else { // space 
			// find the nearest NEC space time
			for (uint8_t space = 1; space < COUNT_OF(SPACES); space++) {
				if (burst[i] > ((SPACES[space-1] + SPACES[space]) / 2)) {
					nec[i] = space;
				} else {
					break;
				}
			}
		}
	}
}

struct nec nec2data(const uint16_t* burst, uint8_t pulses)
{
	uint8_t nec[pulses]; // to store the NEC pulses
	burst2nec(burst, pulses, nec); // convert times to nec pulses
	struct nec to_return;
	to_return.valid = false;
	
	if (pulses > 2+2*8*4) { // numbers of pulses for a normal code
		to_return.repeat = false;
		if (1 == nec[0] && 3 == nec[1] && 0 == nec[66]) { // mark start, space start, trailing mark bit
			uint32_t bits = 0; // all bits
			to_return.valid = true;
			for (uint8_t i = 2; i < 2+2*8*4; i += 2) {
				if (0 == nec[i] && 0 == nec[i+1]) { // mark bit, space 0
					bits = (bits<<1) + 0;
				} else if (0 == nec[i] && 1 == nec[i+1]) { // mark bit, space 1
					bits = (bits<<1) + 1;
				} else { // not a valid bit
					to_return.valid = false;
					break;
				}
			}
			/* decode the valid data into address and command */
			if (to_return.valid) {
				uint8_t address = (bits >> 24) & 0xff; /* first byte is the address */
				uint8_t naddress = (bits >> 16) & 0xff; /* second byte is the inverted address */
				uint8_t command = (bits >> 8) & 0xff; /* third byte is the command */
				uint8_t ncommand = (bits >> 0) & 0xff; /* fourth byte is the inverted command */
				if (0xff == (address ^ naddress)) { /* check if the data is not corrupted (the address comes also inverted) */
					to_return.address = address;
				} else {
					to_return.valid = false;
				}
				if (0xff == (command ^ ncommand)) { /* check if the data is not corrupted (the command comes also inverted) */
					to_return.command = command;
				} else {
					to_return.valid = false;
				}
			}
		}
	} else if (pulses > 2) { // repeat code
		if (1 == nec[0] && 2 == nec[1] && 0 == nec[2]) { // mark start, space repeat, mark bit
			to_return.valid = true;
			to_return.repeat = true;
		}
	}
	return to_return;
}
