/* (C) 2018 King Kevin <kingkevin@cuvoodoo.info>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/* This is the main part of the remote torch controller.
 * It handles all peripherals (IR, serial, torch light)
 */
#include <stdint.h> // Standard Integer Types
#include <stdio.h> // Standard IO facilities
#include <stdlib.h> // General utilities
#include <stdbool.h> // Boolean
#include <avr/io.h> // AVR device-specific IO definitions
#include <avr/interrupt.h> // Interrupts
#include <avr/pgmspace.h> // Program Space Utilities
#include <avr/sleep.h> // sleep utilities
#include <avr/eeprom.h> // EEPROM handling

#include "main.h" // own definitions (e.g. pins)
#include "ir_nec.h" // library to decode NEC codes
#ifdef DEBUG
#include "uart.h"
#endif

// global variables
static volatile uint8_t ir; // IR signal state, to detect change
static volatile int8_t pulse = -1; // pulse index within the burst
/* IR pulse times forming a NEC message
 * a pulse is formed by a mark and a space time
 * pulses:
 * - AGC burst
 * - 8-bit address
 * - 8-bit inverted address
 * - 8-bit command
 * - 8-bit inverted command
 * - end
 */
static volatile uint8_t pulses[2*(1+8*4+1)]; // IR pulse times

// flags, set in the interrupts and handled in the main program
static volatile bool ir_flag = false; // to process a burst

/* torch going out flicker-animation
 * - even index = torch off duration
 * - odd index = torch on duration
 * - unit = T1 overflow (~16.4 ms)
 */
static const uint8_t torch_flicker[] = {5, 5, 5, 30, 5, 5, 20, 10, 5, 5, 5};
static volatile bool torch_tick = false; // on when a tick passed

void led_on(void)
{
	PORTB |= (1 << LED);
}

void led_off(void)
{
	PORTB &= ~(1 << LED);
}

void led_toggle(void)
{
	PINB = (1 << LED);
}

/* switch torch on */
static void torch_on(void)
{
	PORTB |= (1 << TORCH);
}

/* switch torch off */
static void torch_off(void)
{
	PORTB &= ~(1 << TORCH);
}

/* ISR for IR input */
ISR(PCINT0_vect) { // PCI0 Interrupt Vector for PCINT[5:0]
	if (ir == (PINB & (1 << IR))) { // IR signal did not change
		return;
	}
	ir = (PINB & (1 << IR)); // remember change
	if (pulse < 0 && 0 == ir) { // start of first pulse of burst detected (a mark is low)
		pulse = 0; // start recording on end of pulse
	} else if (pulse >= 0 && pulse < COUNT_OF(pulses)) { // save pulse, except the first
		pulses[pulse++] = TCNT1; // save pulse time
	}
	TCNT1 = 0; // clear timer 1
}

/* end of IR burst (using a timeout) */
ISR(TIMER1_OVF_vect) { // timer 1 overflow
	torch_tick = true;
	if (pulse > 0) { // warm a burst is ready to be processed
		ir_flag = true;
	}
}

/* initialize I/O pins */
void ioinit(void)
{
	// configure LED
	DDRB |= (1 << LED); // LED is output
	led_off(); // switch LED off

	// configure torch
	DDRB |= (1 << TORCH); // torch is output
	if (eeprom_read_byte(0)) { // torch state is saved in EEPROM
		torch_off(); // switch torch off
	} else {
		torch_on(); // switch torch off
	}

	// configure InfraRed receiver
	DDRB &= ~(1 << IR); // IR (PB0/PCINT0) receiver is input
	PORTB |= (1 << IR); // enable pull-up to reduce noise
	ir = PINB&(1 << IR); // save state
	GIFR = (1 << PCIF); // clear interrupt flag
	GIMSK |= (1 << PCIE); // enable interrupt for PCINT[5:0]
	PCMSK |= (1 << IR); // enable interrupt for IR input

	// use timer 1 to measure IR pulses
	TCCR1 = (0 << CTC1) | // use FF as TOP (not OCR1C)
	        (0 << PWM1A) | // use CTC mode (not PWM)
	        (0 << COM1A1) | (0 << COM1A0) | // don't use comparator A
	        (0 << CS13) | (1 << CS12) | (1 << CS11) | (1 << CS10); // use CK/64 prescaler (1.0/(1E6/64)*256 = 16.4 ms, this is the most precise prescaler to measure the longest 9 ms pulse, and short enough to overflow before the next code at least 24 ms away)
	TIFR = (1 << TOV1); // clear overflow flag
	TIMSK = (1 << TOIE1); // enable overflow interrupt

#ifdef DEBUG
	// use UART as debug ouput
	uart_init();
#endif

	sei(); // enable interrupts
}

/* action to perform depending on IR code
 * @param[in] address IR code address
 * @param[in] command IR code command
 */
static void ir_action(uint8_t address, uint8_t command)
{
#ifdef DEBUG
	uart_putstr_P(PSTR("nec: a="));
	uart_puthex(address);
	uart_putstr_P(PSTR(" c="));
	uart_puthex(command);
	uart_putchar('\n');
#endif

	if (eeprom_read_byte(0)) {
		torch_on();
		eeprom_update_byte(0, 0x00); // remember we switched on
	} else {
		eeprom_update_byte(0, 0xff); // remember we switched off
		for (uint8_t i = 0; i < COUNT_OF(torch_flicker); i++) {
			if (i % 2) {
				torch_on();
			} else {
				torch_off();
			}
			for (uint8_t ticks = torch_flicker[i]; ticks > 0; ticks--) {
				torch_tick = false;
				while (!torch_tick); // wait until tick passed
			}
		}
		torch_off();
	}
}

int main(void)
{
	ioinit(); // initialize IOs
	set_sleep_mode(SLEEP_MODE_IDLE); // only use idle sleep to keep timer 1 running

	// last IR code
	struct nec last_code = {
		.valid = false,
		.repeat = false,
		.address = 0,
		.command = 0,
	};

#ifdef DEBUG
	uart_putstr_P(PSTR("\n\ntorch controller using IR remote\n"));
#endif

	led_on();
	while (true) {
		while (ir_flag) { // handle IR input
			ir_flag = false; // clear flag
			// check if at least the AGC is present
			if (!(pulse > 2 && pulses[0] > (5000/64))) {
				pulse = -1;
				continue;
			}
			led_off(); // signal IR code has been receive
			// calculate times is us
			uint16_t burst[pulse];
			for (uint8_t i = 0; i < pulse; i++) {
				burst[i] = pulses[i]*(uint16_t)64; // 1s/(F_CPU/T1_PRESCALE) = 1.0/(1E6/64) = 64 us
			}
			// decode burst
			struct nec code = nec2data(burst, pulse); // decode NEC burst
			if (code.valid) {
				// save code
				if (code.repeat) {
					last_code.repeat = true;
				} else {
					last_code = code;
				}
				if (!last_code.repeat) { // ignore repeats
					ir_action(last_code.address, last_code.command); // perform corresponding action
				} else {
#ifdef DEBUG
				uart_putchar('r');
#endif
				}
			} else {
#ifdef DEBUG
				uart_putchar('x');
#endif
			}
			pulse = -1; // reset burst
			led_on(); // IR signal has been processed
		}
		sleep_mode(); // got to sleep
	}
	return 0;
}

